/*
 * License MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Gesso, an extension for Mozilla Firefox
 *
 * The Initial Developer of the Original Code is Ningjie "Jim" Chen.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the License, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the License, the GPL or the LGPL.
 *
 * See MPL.txt for terms of the Mozilla Public License Version 1.1
 * See GPL.txt for terms of the GNU General Public License Version 2.0
 * See LGPL.txt for terms of the GNU Lesser General Public License Version 2.1
 */

var vGesso, vGessoAccServ, vGessoEditSession;
const GUID_PROP_INPUTSCOPE = Components.ID("{1713dd5a-68e7-4a5b-9af6-592a595c778d}");


function GessoDebug(msg) {  
    var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
    consoleService.logStringMessage(msg);
}

/**************************************************
 * 
 *  Some assumptions for GessoTextDocument:
 *   n, this.node = any interface that has an editor
 *       ex. XUL textbox, HTML input("text"), HTML textarea
 *   this.editor = nsIEditor that inherits from nsIPlaintextEditor, but NOT nsIHTMLEditor
 *   this.editor.rootElement = one HTML element (ex. div) that is the only child of its parent
 *       and this element can have zero or more TEXT node(s) as its child(ren). 
 *       no other node type can be its child, except maybe <br>
 *       although <br> shouldn't exist in a "pre" style node (?)
 *   
 */
function GessoTextDocument(n, d, e, evt, win, isl) {
    this.node = n;
    this.doc = d;
    this.editor = e.QueryInterface(Components.interfaces.nsIPlaintextEditor);
    this.root = e.rootElement;
    this.sel = e.selectionController.getSelection(1);
    this.sink = vGesso.addDocument(this);
    this.editing = false;
    
    this.eventHandler = function(o) {return function(e) {return o.handleEvent(e);};}(this);
    win.addEventListener("beforeunload", this.eventHandler, false);
    win.addEventListener("unload", this.eventHandler, false);
    this.eventHost = evt;
    evt.addEventListener("focus", this.eventHandler, true);
    evt.addEventListener("blur", this.eventHandler, true);
    
    this.sel.QueryInterface(Components.interfaces.nsISelectionPrivate).addSelectionListener(this);
    e.addEditActionListener(this);/**/
    
    this.scope = vGesso.comCreateInputScope(this);
    this.scopelist = isl ? isl : new Array();
    this.scopelist.push(vGesso.IS_DEFAULT);
    this.focused = true;
    this.sink.focus();
}
GessoTextDocument.prototype = {
    lock: function(locking) {
    },
    getStatus: function(stat) {
        stat.dynamicFlags = this.editor.isDocumentEditable ? 0 : stat.TS_SD_READONLY;
        stat.staticFlags = stat.TS_SS_NOHIDDENTEXT;
    },
    queryInsert: function(acpTestStart, acpTestEnd, cch, pacpResultStart, pacpResultEnd) {
        if (acpTestStart < 0 || acpTestStart > acpTestEnd || acpTestEnd > this.editor.textLength)
            throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
        var end = acpTestStart + cch;
        pacpResultEnd.value = end;
        pacpResultStart.value = end;
    },
    getSelection: function(ulIndex, pSelection) {
        if (ulIndex == 0) {
            var localsel = this.sel, ancNode = localsel.anchorNode, focNode = localsel.focusNode;
            if (!ancNode || !focNode) throw Components.Exception("no selection", vGesso.TS_E_NOSELECTION);
            var anchor = vGesso.docGetACP(ancNode, localsel.anchorOffset, this.root),
                focus = vGesso.docGetACP(focNode, localsel.focusOffset, this.root);
            if (anchor < focus) {
                pSelection.start = anchor;
                pSelection.end = focus;
                pSelection.selend = pSelection.TS_AE_END;
            }
            else {
                pSelection.start = focus;
                pSelection.end = anchor;
                pSelection.selend = anchor > focus ? pSelection.TS_AE_START : pSelection.TS_AE_NONE;
            }
        }
    },
    getSelectionCount: function(ulCount) {
        ulCount.value = 1;
    },
    setSelection: function(ulIndex, pSelection) {
        if (pSelection.start < 0 || pSelection.start > pSelection.end || pSelection.end > this.editor.textLength)
            throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
        var wasediting = this.editing; this.editing = true;
        var obj = new Object();
        try {
            var sst = pSelection.start, sed = pSelection.end;
            var stnode = vGesso.docGetNodeOffset(sst, this.root, obj),
                stoffset = obj.value;
            var ednode = sed != sst ? vGesso.docGetNodeOffset(sed, this.root, obj) : stnode,
                edoffset = sed != sst ? obj.value : stoffset,
                localsel = this.sel;
            if (pSelection.selstart == pSelection.TS_AE_END) {
                localsel.collapse(stnode, stoffset);
                localsel.extend(ednode, edoffset);
            } else if (sst != sed) {
                localsel.collapse(ednode, edoffset);
                localsel.extend(stnode, stoffset);
            } else
                localsel.collapse(stnode, stoffset);
        } catch (e) {
            throw Components.results.NS_ERROR_FAILURE;
        } finally {
            delete obj;
            this.editing = wasediting;
        }
    },
    getText: function(acpStart, acpEnd, pchPlain, wantPlain, prgRunInfo, ulRunInfoReq, pulRunInfoOut) {
        if (acpEnd == -1) acpEnd = this.editor.textLength;
        if (acpStart < 0 || acpStart > acpEnd || acpEnd > this.editor.textLength)
            throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
        if (wantPlain) {
            var obj = new Object();
            try {
                var stnode = vGesso.docGetNodeOffset(acpStart, this.root, obj), stoffset = obj.value;
                var ednode = vGesso.docGetNodeOffset(acpEnd, this.root, obj);
                var range = this.doc.createRange();
                range.setStart(stnode, stoffset);
                range.setEnd(ednode, obj.value);
                pchPlain.value = range.toString();
                delete obj;
            } catch (e) {
                delete obj;
            }
        }
        if (ulRunInfoReq > 0) {
            var run = prgRunInfo[0];
            run.count = acpEnd - acpStart;
            run.type = run.TS_RT_PLAIN;
            pulRunInfoOut.value = 1;
        }
    },
    getFormattedText: function(ulIndex, wFormat, pchFormattedText) {
        if (ulIndex == 0) {
            wFormat.value = 13; // CF_UNICODETEXT
            this.getText(this.formatStart, this.formatEnd, pchFormattedText, true, null, 0, null, 0);
        }
    },
    getFormattedTextCount: function(acpStart, acpEnd, ulCount) {
        if (acpStart < 0 || acpStart > acpEnd || acpEnd > this.editor.textLength)
            throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
        this.formatStart = acpStart;
        this.formatEnd = acpEnd;
        ulCount.value = 1;
    },
    getEndACP: function(pacp) {
        pacp.value = this.editor.textLength;
    },
    requestSupportedAttrs: function(dwFlags, ulFilterAttrs, paFilterAttrs) {
        for (var i = 0; i < ulFilterAttrs; ++i)
            if (paFilterAttrs[i].value.equals(GUID_PROP_INPUTSCOPE))
                this.wantScope = (dwFlags & vGesso.TS_ATTR_FIND_WANT_VALUE) != 0 ? 1 : 1;
    },
    requestAttrsAtPosition: function(acpPos, ulFilterAttrs, paFilterAttrs, dwFlags) {
        
    },
    requestAttrsTransitioningAtPosition: function(acpPos, ulFilterAttrs, paFilterAttrs, dwFlags) {
        
    },
    findNextAttrTransition: function(acpStart, acpHalt, ulFiltrAettrs, paFilterAttrs, dwFlags, pacpNext, pfFound, plFoundOffset) {
        pacpNext.value = acpHalt; pfFound.value = false;
    },
    retrieveRequestedAttrs: function(ulCount, paAttrVals, pcFetched) {
        var count = 0;
        if (count < ulCount && this.wantScope && this.wantScope != 0) {
            paAttrVals[count].id = GUID_PROP_INPUTSCOPE;
            if (this.wantScope > 0) paAttrVals[count++].objVal = this.scope;
            else paAttrVals[count++].setEmpty();
            delete this.wantScope;
        }
        pcFetched.value = count;
    },
    getActiveView: function(pvcView) {
        pvcView.value = 1;
    },
    getACPFromPoint: function(vcView, pt, dwFlags, pacp) {
        if (vcView == 1) {
            if (acpStart < 0 || acpStart > acpEnd || acpEnd > this.editor.textLength)
                throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
            var clip = new Object(), text, acc;
            try {
                text = this.editor.rootElement.firstChild;
                if (text.nodeType == 3) acc = vGessoAccServ.getAccessibleFor(text); 
                while (text.nextSibling != null) {
                    if (text.nodeType == 3) {
                        var rc = vGesso.accTextGetBounds(acc, 0, text.length, clip);
                        if (pt.y < rc.top || (pt.y < rc.bottom && pt.x < rc.right)) break;
                    }
                    text = text.nextSibling;
                    if (text.nodeType == 3) acc = vGessoAccServ.getAccessibleFor(text); 
                }
                delete clip;
            } catch (e) {
                pacp.value = 0;
                delete clip;
                return;
            }
            if (text && acc)
                pacp.value = vGesso.docGetACP(text, vGesso.accTextGetOffset(acc, pt.x, pt.y, dwFlags, text.length), this.root);
        }
    },
    // a nice little hack to make the day a little brighter, just a little
    //    - ISimpleDOMText has trouble returning bounds if the indices are the same.
    getTextExt: function(vcView, acpStart, acpEnd, prc, pfClipped) {
        if (vcView == 1) {
            var len = this.editor.textLength;
            if (len > 0) {
                if (acpStart < 0 || acpStart > acpEnd || acpEnd > len)
                    throw Components.Exception("invalid position", vGesso.TS_E_INVALIDPOS);
                var obj = new Object();
                try {
                    var l = Number.MAX_VALUE, t = l, r = -l, b = r, clip = false;
                    var startnode = vGesso.docGetTextNodeOffset(acpStart, this.root, obj), 
                        startoffset = obj.value;
                    var endnode = vGesso.docGetTextNodeOffset(acpEnd, this.root, obj),
                        endoffset = obj.value;
                    if (endnode == startnode && endoffset == startoffset) 
                        if (startoffset > 0) --startoffset;
                        else ++endoffset;
                    for (;;) {
                        var ed = startnode == endnode ? endoffset : startnode.length - 1; 
                        var rc = vGesso.accTextGetBounds(vGessoAccServ.getAccessibleFor(startnode), startoffset, ed, obj);
                        l = Math.min(l, rc.left); t = Math.min(t, rc.top);
                        r = Math.max(r, rc.right); b = Math.max(b, rc.bottom);
                        clip |= obj.value;
                        if (startnode == endnode) break;
                        startnode = startnode.nextSibling; startoffset = 0;
                    }
                    delete obj;
                    if (l == r && t == b) { // text measuring quirk
                        prc.left = this.extl; prc.right = this.extr;
                        prc.top = this.extt; prc.bottom = this.extb;
                    } else {
                        prc.left = this.extl = l; prc.right = this.extr = r; 
                        prc.top = this.extt = t; prc.bottom = this.extb = b;
                    }
                    pfClipped.value = clip;
                    //GessoDebug("getTextExt: " + acpStart + "-" + acpEnd + ":(" + l + "," + t + "-" + r + "," + b + ")");
                    return;
                } catch (e) {
                    delete obj;
                }
            }
            var box = this.doc.getBoxObjectFor(this.node);
            prc.right = (prc.left = box.screenX) + 1;
            prc.bottom = Math.min((prc.top = box.screenY) + 24, box.screenY + box.height);
            pfClipped.value = false;
        }
    },
    getScreenExt: function(vcView, prc) {
        if (vcView == 1) {
            var box = this.doc.getBoxObjectFor(this.node);
            var x = box.screenX; prc.left = x; prc.right = x + box.width;
            var y = box.screenY; prc.top = y; prc.bottom = y + box.height; 
        }
    },
    getAccessNode: function(vcView) {
        if (vcView == 1)
            try {
                var n = this.eventHost;
                if (n.accessible) return n.accessible.QueryInterface(Components.interfaces.nsIAccessNode);
                else return vGessoAccServ.getAccessibleFor(n).QueryInterface(Components.interfaces.nsIAccessNode);
            } catch (e) {
                throw Components.results.NS_ERROR_FAILURE;
            }
    },
    // fixed insertText() doesn't work with zero-length strings
    //    and deleteFromDocument() doesn't work if there's no selection, makes sense right?
    //        until you realize it doesn't just quit but instead throws at you a nasty exception.
    // *REMOVED* insertText() also messes up the selection sometimes when we join the nodes in 
    //  DidInsertNode. therefore we manually fix the selection
    insertTextAtSelection: function(dwFlags, pchText, pacpStart, pacpEnd, pChange) {
        var wasediting = this.editing; this.editing = true;
        var obj = new Object();
        try {
            var localsel = this.sel;
            var st = vGesso.docGetACP(localsel.anchorNode, localsel.anchorOffset, this.root);
            var ed = vGesso.docGetACP(localsel.focusNode, localsel.focusOffset, this.root);
            if (st > ed) {var tmp = ed; ed = st; st = tmp;}
            if (dwFlags & vGesso.TS_IAS_QUERYONLY) {
                ed = st + pchText.length;
                pacpStart.value = ed;
                pacpEnd.value = ed;
                if (!pChange.isNull) {
                    pChange.start = st;
                    pChange.oldEnd = ed; 
                    pChange.newEnd = st + pchText.length;
                }
            } else {
                pChange.start = st; pChange.oldEnd = ed;
                if (pchText.length > 0) {
                    this.editor.insertText(pchText);
                    ed = st + pchText.length;
                    //var selnode = vGesso.docGetNodeOffset(ed, this.root, obj);
                    //localsel.collapse(selnode, obj.value);
                }
                else if (st != ed) {
                    localsel.deleteFromDocument();
                    ed = st;
                }
                pChange.newEnd = ed;
                if ((dwFlags & vGesso.TF_IAS_NOQUERY) == 0) {
                    pacpStart.value = vGesso.docGetACP(localsel.anchorNode, localsel.anchorOffset, this.root);
                    pacpEnd.value = ed;
                }
            }
        } catch (e) {
            throw Components.results.NS_ERROR_FAILURE;
        } finally {
            delete obj;
            this.editing = wasediting;
        }
    }, 
    getInputScopes: function(count) {
        count.value = this.scopelist.length;
        return this.scopelist;
    },
    getPhrase: function(count) {
        count.value = 0;
    },
    getRegularExpression: function() {
    },
    getSRGS: function() {
    },
    getXML: function() {
    },
    handleEvent: function(e) {
        var targ = e.currentTarget;
        switch (e.type) {
        case "beforeunload":
        case "unload":
            delete this.focused;
            delete this.scope;
            delete this.scopelist;
            this.editor.removeEditActionListener(this);
            this.sel.QueryInterface(Components.interfaces.nsISelectionPrivate).removeSelectionListener(this);
            
            this.eventHost.removeEventListener("blur", this.eventHandler, true);
            this.eventHost.removeEventListener("focus", this.eventHandler, true);
            delete this.eventHost;
            targ.removeEventListener("beforeunload", this.eventHandler, false);
            targ.removeEventListener("unload", this.eventHandler, false);
            delete this.eventHandler;
            
            delete this.editing;
            this.sink.removeDocument();
            delete this.sink;
            delete this.sel;
            delete this.root;
            delete this.editor;
            delete this.doc;
            GessoDebug("Gesso: " + this.node.tagName + " unloaded");
            delete this.node;
            delete this;
            break;
        case "focus":
            if (!this.focused) {
                this.sink.focus();
                this.focused = true;
            }
            break;
        case "blur":
            if (this.focused) {
                this.sink.unfocus();
                this.focused = false;
            }
            break;
        }
    },
    DeleteWorker: function() {
        if (!this.editing) {
            var st = this.editStart, ed = this.editEnd;
            this.sink.onTextChange(0, {start: st, oldEnd: ed, newEnd: st});
        }
    },
    notifySelectionChanged: function(doc, sel, reason) {
        if (!this.editing) {
            this.sink.onSelectionChange();
        }
    },
    DidCreateNode: function (tag, node, parent, pos, res) {
        this.DidInsertNode(node, parent, pos, res);
    },
    DidDeleteNode: function (child, res) {
        this.DeleteWorker();
    },
    DidDeleteSelection: function(sel) {
        this.DeleteWorker();
    },
    DidDeleteText: function(node, offset, len, res) {
        this.DeleteWorker();
    },
    // NEEDS MORE RESEARCH (GetACP and GetNodeText are native now):
    // insertText has this habit of creating extra text nodes
    // try fixing it here, otherwise GetACP and GetNodeText will become slower and slower
    DidInsertNode: function(node, parent, pos, res) {
        /*
        if (!this.editing) {
            if (pos >= parent.childNodes.length) pos = parent.childNodes.length - 1;
            var st = vGesso.docGetACP(parent, pos, this.root), 
                ed = vGesso.docGetACP(parent, pos + 1, this.root); 
            this.sink.onTextChange(0, {start: st, oldEnd: st, newEnd: ed});
        } else if (node.nodeType == 3) {
            var wasediting = this.editing; this.editing = true;
            var prevsib = node.previousSibling, prevjoin = prevsib ? prevsib.nodeType == 3 : false,
                nextsib = node.nextSibling, nextjoin = nextsib ? nextsib.nodeType == 3 : false;
            if (prevjoin) this.editor.joinNodes(prevsib, node, parent);
            if (nextjoin) this.editor.joinNodes(node, nextsib, parent);
            this.editing = wasediting;
        }
        */
    },
    DidInsertText: function(node, offset, str, res) {
        if (!this.editing) {
            var st = vGesso.docGetACP(node, offset, this.root), ed = st + str.length;
            this.sink.onTextChange(0, {start: st, oldEnd: st, newEnd: ed});
        }
    },
    DidJoinNodes: function(left, right, parent, res) {
    },
    DidSplitNode: function(right, offset, left, res) {
    },
    WillCreateNode: function(tag, parent, pos) {
    },
    WillDeleteNode: function(child) {
        if (!this.editing) {
            this.editStart = vGesso.docGetACP(child, 0, this.root);
            this.editEnd = this.editStart + child.length;
        }
    },
    WillDeleteSelection: function(sel) {
        if (!this.editing) {
            this.editStart = vGesso.docGetACP(sel.anchorNode, sel.anchorOffset, this.root);
            this.editEnd = vGesso.docGetACP(sel.focusNode, sel.focusOffset, this.root);
        }
    },
    WillDeleteText: function(node, offset, len) {
        if (!this.editing) {
            this.editStart = vGesso.docGetACP(node, offset, this.root);
            this.editEnd = this.editStart + len;
        }
    },
    WillInsertNode: function(node, parent, pos) {
    },
    WillInsertText: function(node, offset, str) {
    },
    WillJoinNodes: function(left, right, parent) {
    },
    WillSplitNode: function(right, offset) {
    }
};

function GessoWindowFocus(e) {
    var disp = e.currentTarget.document.commandDispatcher;
    var targ = e.target;
    if (!targ) targ = disp.focusedElement;
    var targwin = targ && targ.ownerDocument ? targ.ownerDocument.defaultView : null;
    if (!targwin) targwin = disp.focusedWindow;
    if (targwin) {
        var design = targwin.document.designMode && (targwin.document.designMode.toLowerCase() == "on");
        if (design ? !targwin.gessoSearched : targ && !targ.gessoSearched) {
            try {
                if (design) {
                    targwin.gessoSearched = true;
                    var editor = vGessoEditSession.getEditorForWindow(targwin);
                    new GessoTextDocument(targwin.document.documentElement, targwin.document, editor, targwin.document, targwin, null);
                    GessoDebug("Gesso: rich editor (experimental support) added.");
                } else {
                    targ.gessoSearched = true;
                    
                    var isl = new Array();
                    if (targ.id && targ.id == "urlbar")
                        isl.push(vGesso.IS_URL);
                    else if ((targ.id && (targ.id.toLowerCase() == "username" || targ.id.toLowerCase() == "login")) ||
                        (targ.name && (targ.name.toLowerCase() == "username" || targ.name.toLowerCase() == "login")))
                        isl.push(vGesso.IS_LOGINNAME);
                    else if ((targ.id && targ.id.toLowerCase() == "email") ||
                        (targ.name && targ.name.toLowerCase() == "email"))
                        isl.push(vGesso.IS_EMAIL_SMTPEMAILADDRESS);
                    
                    if (targ.textbox) targ = targ.textbox; // searchbar hack
                    if (!targ.editor) targ = targ.QueryInterface(Components.interfaces.nsIDOMNSEditableElement);
                    
                    new GessoTextDocument(targ, targ.ownerDocument, targ.editor, targ, targwin, isl);
                    GessoDebug("Gesso: " + targ.tagName + " added.");
                }
            } catch (e) {
            }
        }
    }
}

function GessoWindowUnload(e) {
    e.target.removeEventListener("focus", GessoWindowFocus, true);
    e.target.removeEventListener("unload", GessoWindowUnload, false);
}

vGesso = Components.classes["@gesso.mozdev.org/gesso;1"].getService(Components.interfaces.iGesso);
vGessoAccServ = Components.classes["@mozilla.org/accessibilityService;1"].getService(Components.interfaces.nsIAccessibleRetrieval);
vGessoEditSession = Components.classes["@mozilla.org/editor/editingsession;1"].getService(Components.interfaces.nsIEditingSession);
this.addEventListener("focus", GessoWindowFocus, true);
this.addEventListener("unload", GessoWindowUnload, false);

